
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

#ifndef	Float
#define Float	float
#endif

#ifndef	Filter
#define Filter	Ramachandran
#endif

#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif

extern void	Error(char *msg);
extern double	Filter(int j);

#ifdef	_WIN32
#include <process.h>
#include <windows.h>

#define THREAD_T	HANDLE
#define FUNCTION_T	unsigned
#define FUNCTION_T_CC	FUNCTION_T __stdcall
#define RETURN_VALUE	0

#define INIT_THREAD(T,F,A)	\
	(T=(HANDLE)_beginthreadex(NULL,0,F,(void *)A,0,NULL))==0
#define TERM_THREAD(T)	\
	WaitForSingleObject(T,INFINITE)==WAIT_FAILED || CloseHandle(T)==0
#else
#include <pthread.h>

#define THREAD_T	pthread_t
#define FUNCTION_T	void *
#define FUNCTION_T_CC	FUNCTION_T
#define RETURN_VALUE	NULL

#define INIT_THREAD(T,F,A)	pthread_create(&(T),NULL,F,(void *)A)
#define TERM_THREAD(T)		pthread_join(T,NULL)
#endif

static int	Threads=1;
static THREAD_T	*Thread;

static void	MultiThreading(FUNCTION_T (*body)(),char *name)
{
	size_t	i;

	for (i=1; i<Threads; i++)
	    if (INIT_THREAD(Thread[i],body,i)) {
		(void)fputs(name,stderr);
		Error(" : multi-threading initialization error.");
	    }

	(void)(FUNCTION_T_CC)(*body)((void *)0);

	for (i=1; i<Threads; i++)
	    if (TERM_THREAD(Thread[i])) {
		(void)fputs(name,stderr);
		Error(" : multi-threading termination error.");
	    }
}

#define MULTI_THREADING(FUNCTION)	MultiThreading(FUNCTION,#FUNCTION)

typedef struct {
		double	x,y;
	} Vector;

typedef struct {
		int	j;
		double	Eu,F4l,F4F;
	} Struct;

static int	Nu,Nw,N1,N2,O,*C;
static double	Ow,**F4Q,**PQ;
static Vector	*E,*G,**V;
static Struct	**S4BP;
static Float	***F;

double	Ramachandran(int j)
{
	return (j==0)?1.0/4.0:(j&1)?-1.0/(M_PI*M_PI*(double)j*(double)j):0.0;
}

double	Shepp(int j)
{
	return 2.0/(M_PI*M_PI*(1.0-4.0*(double)j*(double)j));
}

double	Chesler(int j)
{
	if (j==0)
	    return 1.0/8.0-1.0/(2.0*M_PI*M_PI);
	else if (j==1 || j==(-1))
	    return 1.0/16.0-1.0/(2.0*M_PI*M_PI);
	else if (j&1)
	    return -1.0/(2.0*M_PI*M_PI*(double)j*(double)j);
{
	double  j2=(double)j*(double)j,j21=j2-1.0;

	return -(j2+1.0)/(2.0*M_PI*M_PI*j21*j21);
}}

static void	FFT(int sign,Vector *GV)
{
	int	j,k,l,m,p,q;
	Vector	gv,*e,*a,*b;
	double	x,y,z;

	for (j=N2-1, k=N2-2; k>0; k--) {
	    for (l=N1; (j^=l)&l; l>>=1) ;

	    if (j>k) {
		gv=GV[j]; GV[j]=GV[k]; GV[k]=gv;
	    }
	}
	for (j=2, k=1, l=N1, m=N2; m>1; m=l, l>>=1, k=j, j<<=1)
	    for (e=E, p=0; p<k; p++, e+=m) {
		x=e->x; y=e->y*(double)sign;
		for (a=GV+p, q=0; q<l; q++, a+=j) {
		    gv=(*(b=a+k)); b->x=a->x-(z=gv.x*x-gv.y*y); a->x+=z;
				   b->y=a->y-(z=gv.x*y+gv.y*x); a->y+=z;
		}
	    }
}

static FUNCTION_T_CC	Convolution(void *i)
{
	size_t	l;
	double	*f4q,*pq,x,y;
	int	k,j;
	Vector	*v=V[(size_t)i];

	for (l=(size_t)i; l<Nw; l+=Threads) {
	    f4q=F4Q[l]; pq=PQ[l];

	    for (k=0; k<N1; k++) v[k].x=v[k].y=0.0;

	    for (j=0; j<Nu; j++, k++) {
		v[k].x=f4q[j]*pq[j]; v[k].y=0.0;
	    }
	    for (; k<N2; k++) v[k].x=v[k].y=0.0;

	    FFT(-1,v);

	    for (k=0; k<N2; k++) {
		x=v[k].x;
		y=v[k].y; v[k].x=G[k].x*x-G[k].y*y;
			  v[k].y=G[k].x*y+G[k].y*x;
	    }
	    FFT(1,v);

	    for (j=0; j<Nu; j++) pq[j]=v[j].x;
	}
	return RETURN_VALUE;
}

static FUNCTION_T_CC	BackProjection(void *i)
{
	size_t	z;
	double	Z,u,wl,w,Q,Qu,Qw;
	int	c,S,y,Y,s,x,X,j1,j,l1,l;
	Float	**Fz,*Fzy;
	Struct	*s4bp;

	for (z=(size_t)i; z<Nw; z+=Threads) {
	    Z=(double)z-Ow; c=C[z]; S=c*c; Fz=F[z];
	    for (y=O+(Y=(-c)); Y<=c; Y++, y++) {
		s=S-Y*Y; s4bp=S4BP[y]; Fzy=Fz[y];
		for (x=O+(X=(-c)); X<=c; X++, x++)
		    if (s-X*X>=0)
{
	u=s4bp[x].Eu; j1=(j=s4bp[x].j)+1;
	w=modf(Ow+s4bp[x].F4l*Z,&wl); l1=(l=(int)wl)+1;
	Q=PQ[l][j]; Qu=PQ[l][j1]-Q; Qw=PQ[l1][j]-Q;
	Fzy[x]+=(s4bp[x].F4F*(Q+Qu*u+Qw*w+(PQ[l1][j1]-Q-Qu-Qw)*u*w));
}
	    }
	}
	return RETURN_VALUE;
}

#define LEN	2048

int	main(int argc,char **argv)
{
	char	*env,str[LEN],form[LEN],range[LEN],slice[LEN];
	FILE	*file;
	int	Nt,BPS,N,j,i,l,z,y,x,m,Y,X,lim,norm=0;
	double	A,B,Du,Ou,Dw,u,o,t0,base,step,w,s,c,F4l,uj,roof,min,max;
	Cell	*cell;

	if (argc!=11 && argc!=13)
	    Error("usage : xp2tg XP A B Du Ou Dw Ow t0 {base step} BPS TG/");

	if ((env=getenv("THREADS"))!=NULL &&
	    (Threads=atoi(env))<=0) Error("bad number of THREADS.");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (fgets(str,LEN,file)==NULL ||
	    sscanf(str,"%d %d %d",&Nu,&Nw,&Nt)!=3 ||
	    Nu<=0 || Nw<=0 || Nt<=0) Error("bad projection data size.");

	if ((A=atof(argv[2]))<=0.0) Error("bad source-sample distance.");
	if ((B=atof(argv[3]))<A) Error("bad source-detector distance.");

	if ((Du=atof(argv[4]))<=0.0 ||
	    (Dw=atof(argv[6]))<=0.0) Error("bad inerval of dectors.");

	if ((Ou=atof(argv[5]))<=0.0 || Ou>=(double)(Nu-1) ||
	    (Ow=atof(argv[7]))<=0.0 || Ow>=(double)(Nw-1))
	    Error("bad center position of detector array.");

	u=Du*Ou/B;
	o=Ou/sqrt(1.0+u*u);
	u=Du*((double)(Nu-1)-Ou)/B;
	if ((u=((double)(Nu-1)-Ou)/sqrt(1.0+u*u))<o) o=u;

	if (Du*o>=B) Error("bad geometric configuration.");

	t0=atof(argv[8])*M_PI/180.0;

	BPS=atoi(argv[argc-2]);

	if (argc==13) {
	    base=atof(argv[9]);
	    if ((step=atof(argv[10]))<=0.0) Error("bad CT-value step.");
	}
	else
	    if ((norm=BPS<0)) BPS=(-BPS);

	if (BPS<=0 || BPS>sizeof(Cell)*8) Error("bad BPS of CT image.");

	for (N1=1; N1<Nu; N1<<=1) ; N2=N1<<1;

	N=(O=(int)floor(o))*2+1;

	if ((Thread=(THREAD_T *)malloc(sizeof(THREAD_T)*Threads))==NULL ||
	    (E=(Vector *)malloc(sizeof(Vector)*N2))==NULL ||
	    (G=(Vector *)malloc(sizeof(Vector)*N2))==NULL ||
	    (V=(Vector **)malloc(sizeof(Vector *)*Threads))==NULL ||
	    (*V=(Vector *)malloc(sizeof(Vector)*Threads*N2))==NULL ||
	    (F4Q=(double **)malloc(sizeof(double *)*Nw))==NULL ||
	    (*F4Q=(double *)malloc(sizeof(double)*Nw*Nu))==NULL ||
	    (PQ=(double **)malloc(sizeof(double *)*(Nw+1)))==NULL ||
	    (*PQ=(double *)malloc(sizeof(double)*((size_t)Nw*Nu+1)))==NULL ||
	    (C=(int *)malloc(sizeof(int)*Nw))==NULL ||
	    (S4BP=(Struct **)malloc(sizeof(Struct *)*N))==NULL ||
	    (*S4BP=(Struct *)malloc(sizeof(Struct)*N*N))==NULL ||
	    (F=(Float ***)malloc(sizeof(Float **)*Nw))==NULL ||
	    (*F=(Float **)malloc(sizeof(Float *)*Nw*N))==NULL ||
	    (**F=(Float *)malloc(sizeof(Float)*Nw*N*N))==NULL)
	    Error("memory allocation error.");

	for (j=0; j<N2; j++) {
	    sincos(M_PI*(double)j/(double)N2,&(E[j].y),&(E[j].x));

	    G[j].x=Filter(j-N1); G[j].y=0.0;
	}
	FFT(-1,G);

	for (i=1; i<Threads; i++) V[i]=V[i-1]+N2;

	for (l=0; l<Nw; l++) {
	    if (l>0) {
		F4Q[l]=F4Q[l-1]+Nu; PQ[l]=PQ[l-1]+Nu;
	    }
	    w=Dw*((double)l-Ow);
	    for (j=0; j<Nu; j++) {
		u=Du*((double)j-Ou);
		F4Q[l][j]=A/Du/sqrt(B*B+u*u+w*w)/(double)N2;
	    }
	}
	PQ[Nw]=PQ[0];

	l=(int)ceil(o/(w=B/Du/Ow));
	for (z=0; z<l; z++) C[z]=(int)floor(w*(double)z);

	l=Nw-1-(int)floor(o/(w=B/Du/((double)(Nw-1)-Ow)));
	for (; z<l ; z++) C[z]=O;
	for (; z<Nw; z++) C[z]=(int)floor(w*(double)(Nw-1-z));

	for (y=1; y<N; y++) S4BP[y]=S4BP[y-1]+N;

	for (z=0; z<Nw; z++) {
	    if (z>0) F[z]=F[z-1]+N;

	    for (y=0; y<N; y++) {
		if (z>0 || y>0) F[z][y]=F[z][y-1]+N;

		for (x=0; x<N; x++) F[z][y][x]=0.0;
	    }
	}
	for (m=0; m<Nt; m++) {
	    if (fread(*PQ,sizeof(double),(size_t)Nw*Nu,file)!=(size_t)Nw*Nu)
		Error("projection data read error.");

	    MULTI_THREADING(Convolution);

	    sincos(t0+2.0*M_PI*(double)m/(double)Nt,&s,&c);
	    for (Y=(-O), y=0; y<N; y++, Y++)
	    for (X=(-O), x=0; x<N; x++, X++) {
		S4BP[y][x].F4l=F4l=B/(B+Du*(c*(double)Y-s*(double)X));
		S4BP[y][x].F4F=F4l*F4l*B*B*M_PI/A/A/(double)Nt;
		S4BP[y][x].Eu=modf(Ou+F4l*(c*(double)X+s*(double)Y),&uj);
		S4BP[y][x].j=(int)uj;
	    }
	    MULTI_THREADING(BackProjection);
	}
	if (file!=stdin) (void)fclose(file);

	i=1; for (z=Nw-1; z>=10; z/=10) ++i; (void)sprintf(form,"%%0%dd",i);

	lim=(1<<BPS)-1;

	if (argc==13)
	    (void)sprintf(range,"%lf\t%lf",base,roof=base+step*(double)lim);
	else if (norm) {
	    min=max=F[0][0][0];
	    for (z=0; z<Nw; z++)
	    for (y=0; y<N ; y++)
	    for (x=0; x<N ; x++)
		if (F[z][y][x]<min) min=F[z][y][x]; else
		if (F[z][y][x]>max) max=F[z][y][x];

	    (void)sprintf(range,"%lf\t%lf",min,max);
	    base=min; step=(min<max)?(max-min)/(double)lim:1.0;

	    (void)fputs(range,stderr); (void)fputc('\n',stderr);
	}
	for (z=0; z<Nw; z++) {
	    min=max=F[z][0][0];
	    for (y=0; y<N; y++)
	    for (x=0; x<N; x++)
		if (F[z][y][x]<min) min=F[z][y][x]; else
		if (F[z][y][x]>max) max=F[z][y][x];

	    (void)sprintf(slice,form,z);
	    (void)printf("%s\t%d\t%lf\t%lf\n",slice,C[z],min,max);

	    if (argc==13)
		for (y=0; y<N; y++) {
		    cell=(Cell *)S4BP[N-1-y];
		    for (x=0; x<N; x++)
			cell[x]=(F[z][y][x]<base)?0:
				(F[z][y][x]>roof)?(Cell)lim:
				(Cell)((F[z][y][x]-base)/step+0.5);
		}
	    else {
		if (!norm) {
		    (void)sprintf(range,"%lf\t%lf",min,max);
		    base=min; step=(min<max)?(max-min)/(double)lim:1.0;
		}
		for (y=0; y<N; y++) {
		    cell=(Cell *)S4BP[N-1-y];
		    for (x=0; x<N; x++)
			cell[x]=(Cell)((F[z][y][x]-base)/step+0.5);
		}
	    }
	    (void)sprintf(str,"%s/%s.tif",argv[argc-1],slice);
	    StoreImageFile(str,N,N,BPS,(Cell **)S4BP,range);
	}
	free(**F); free(*F); free(F); free(*S4BP); free(S4BP); free(C);
	free(*PQ); free(PQ); free(*F4Q); free(F4Q);
	free(*V); free(V); free(G); free(E);
	free(Thread);

	return 0;
}
