
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

#ifndef	Float
#define Float	float
#endif

#ifndef	Filter
#define Filter	Ramachandran
#else
extern double	Filter(int j);
#endif

extern void	Error(char *msg);

#ifdef	_WIN32
#include <process.h>
#include <windows.h>

#define THREAD_T	HANDLE
#define FUNCTION_T	unsigned __stdcall
#define RETURN_VALUE	0

#define INIT_THREAD(T,F,A)	\
	(T=(HANDLE)_beginthreadex(NULL,0,F,(void *)A,0,NULL))==0
#define TERM_THREAD(T)	\
	WaitForSingleObject(T,INFINITE)==WAIT_FAILED || CloseHandle(T)==0
#else
#include <pthread.h>

#define THREAD_T	pthread_t
#define FUNCTION_T	void *
#define RETURN_VALUE	NULL

#define INIT_THREAD(T,F,A)	pthread_create(&(T),NULL,F,(void *)A)
#define TERM_THREAD(T)		pthread_join(T,NULL)
#endif

#define INIT_MT(T,F,A)	\
if (INIT_THREAD(T,F,A)) Error(#F " : multi-threading initialization error.")

#define TERM_MT(T,F)	\
	if (TERM_THREAD(T)) Error(#F " : multi-threading termination error.")

#ifndef	M_PI
#define	M_PI	3.14159265358979323846
#endif

double	Ramachandran(int j)
{
	return (j==0)?1.0/4.0:(j&1)?-1.0/(M_PI*M_PI*(double)j*(double)j):0.0;
}

double	Shepp(int j)
{
	return 2.0/(M_PI*M_PI*(1.0-4.0*(double)j*(double)j));
}

double	Chesler(int j)
{
	if (j==0)
	    return 1.0/8.0-1.0/(2.0*M_PI*M_PI);
	else if (j==1 || j==(-1))
	    return 1.0/16.0-1.0/(2.0*M_PI*M_PI);
	else if (j&1)
	    return -1.0/(2.0*M_PI*M_PI*(double)j*(double)j);
{
	double  j2=(double)j*(double)j,j21=j2-1.0;

	return -(j2+1.0)/(2.0*M_PI*M_PI*j21*j21);
}}

typedef struct {
		double	x,y;
	} Vector;

typedef struct {
		int	j;
		double	Eu,F4l,F4Q;
	} Struct;

static int	Nu,Nw,N1,N2,N,Ol,Nl,Threads=1;
static double	Ow,**F4P,**P,**Q;
static Vector	*E,*G,**V;
static Struct	**S4BP;
static Float	***F;

static void	FFT(int sign,Vector *GV)
{
	int	j,k,l,m,p,q;
	Vector	gv,*e,*a,*b;
	double	x,y,z;

	for (j=N2-1, k=N2-2; k>0; k--) {
	    for (l=N1; (j^=l)&l; l>>=1) ;

	    if (j>k) {
		gv=GV[j]; GV[j]=GV[k]; GV[k]=gv;
	    }
	}
	for (j=2, k=1, l=N1, m=N2; m>1; m=l, l>>=1, k=j, j<<=1)
	    for (e=E, p=0; p<k; p++, e+=m) {
		x=e->x; y=e->y*(double)sign;
		for (a=GV+p, q=0; q<l; q++, a+=j) {
		    gv=(*(b=a+k)); b->x=a->x-(z=gv.x*x-gv.y*y); a->x+=z;
				   b->y=a->y-(z=gv.x*y+gv.y*x); a->y+=z;
		}
	    }
}

static FUNCTION_T	Convolution(void *i)
{
	size_t	l;
	int	k,j;
	double	*f4p,*p,x,y;
	Vector	*v=V[(size_t)i];

	for (l=(size_t)i; l<Nw; l+=Threads) {
	    for (k=0; k<N1; k++) v[k].x=v[k].y=0.0;

	    f4p=F4P[l]; p=P[l];
	    for (j=0; j<Nu; j++, k++) {
		v[k].x=f4p[j]*p[j]; v[k].y=0.0;
	    }
	    for (; k<N2; k++) v[k].x=v[k].y=0.0;

	    FFT(-1,v);

	    for (k=0; k<N2; k++) {
		x=v[k].x;
		y=v[k].y; v[k].x=G[k].x*x-G[k].y*y;
			  v[k].y=G[k].x*y+G[k].y*x;
	    }
	    FFT(1,v);

	    for (j=0; j<Nu; j++) Q[j][l]=v[j].x;
	}
	return RETURN_VALUE;
}

static FUNCTION_T	BackProjection(void *i)
{
	size_t	y;
	int	x,j,z,l;
	Float	**Fy,*Fyx;
	Struct	*s4bp;
	double	*Q0,*Q1,u,F4Q,W,F4l,w,q,Qu,Qw,Z=(double)Ol-Ow;

	for (y=(size_t)i; y<N; y+=Threads) {
	    Fy=F[y]; s4bp=S4BP[y];
	    for (x=0; x<N; x++)
		if ((j=s4bp[x].j)>=0) {
		    Fyx=Fy[x];
		    Q0=Q[j]; Q1=Q[j+1];
		    u=s4bp[x].Eu; F4Q=s4bp[x].F4Q;
		    for (W=Ow+(F4l=s4bp[x].F4l)*Z, z=0; z<Nl; z++, W+=F4l) {
			w=W-(double)(l=(int)W);
			Qu=Q1[l]-(q=Q0[l]); Qw=Q0[++l]-q;
			Fyx[z]+=(F4Q*(q+Qu*u+Qw*w+(Q1[l]-q-Qu-Qw)*u*w));
		    }
		}
	}
	return RETURN_VALUE;
}

#define LEN	2048

int	main(int argc,char **argv)
{
	char		*env,str[LEN],form[LEN],range[LEN],slice[LEN];
	FILE		*file;
	int		Nt,BPS,O,j,l,O2,Y,y,O2_Y2,X,x,z,m,lim,norm=0;
	double		A,B,Du,Ou,Dw,u,o,t0,base,step,w,s,c,F4l,roof,min,max;
	THREAD_T	*Thread;
	size_t		i;
	Cell		*cell;

	if (argc!=11 && argc!=13)
	    Error("usage : xp2tg XP A B Du Ou Dw Ow t0 {base step} BPS TG/");

	if ((env=getenv("THREADS"))!=NULL &&
	    (Threads=atoi(env))<=0) Error("bad number of THREADS.");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (fgets(str,LEN,file)==NULL ||
	    sscanf(str,"%d %d %d",&Nu,&Nw,&Nt)!=3 ||
	    Nu<=0 || Nw<=0 || Nt<=0) Error("bad projection data size.");

	if ((A=atof(argv[2]))<=0.0) Error("bad source-sample distance.");
	if ((B=atof(argv[3]))<A) Error("bad source-detector distance.");

	if ((Du=atof(argv[4]))<=0.0 ||
	    (Dw=atof(argv[6]))<=0.0) Error("bad inerval of dectors.");

	if ((Ou=atof(argv[5]))<=0.0 || Ou>=(double)(Nu-1) ||
	    (Ow=atof(argv[7]))<=0.0 || Ow>=(double)(Nw-1))
	    Error("bad center position of detector array.");

	u=Du*Ou/B;
	o=Ou/sqrt(1.0+u*u);
	u=Du*((double)(Nu-1)-Ou)/B;
	if ((u=((double)(Nu-1)-Ou)/sqrt(1.0+u*u))<o) o=u;

	if (o*Du>=B) Error("bad geometric configuration.");

	t0=atof(argv[8])*M_PI/180.0;

	BPS=atoi(argv[argc-2]);

	if (argc==13) {
	    base=atof(argv[9]);
	    if ((step=atof(argv[10]))<=0.0) Error("bad CT-value step.");
	}
	else
	    if ((norm=BPS<0)) BPS=(-BPS);

	if (BPS<=0 || BPS>sizeof(Cell)*8) Error("bad BPS of CT image.");

	for (N1=1; N1<Nu; N1<<=1) ; N2=N1<<1;

	N=(O=(int)floor(o))*2+1;
	Ol=(int)ceil(o*Du*Ow/B);
	Nl=Nw-1-(int)floor(o*Du*((double)(Nw-1)-Ow)/B)-Ol;

	if ((Thread=(THREAD_T *)malloc(sizeof(THREAD_T)*Threads))==NULL ||
	    (E=(Vector *)malloc(sizeof(Vector)*N2))==NULL ||
	    (G=(Vector *)malloc(sizeof(Vector)*N2))==NULL ||
	    (V=(Vector **)malloc(sizeof(Vector *)*Threads))==NULL ||
	    (*V=(Vector *)malloc(sizeof(Vector)*Threads*N2))==NULL ||
	    (F4P=(double **)malloc(sizeof(double *)*Nw))==NULL ||
	    (*F4P=(double *)malloc(sizeof(double)*Nw*Nu))==NULL ||
	    (P=(double **)malloc(sizeof(double *)*Nw))==NULL ||
	    (*P=(double *)malloc(sizeof(double)*Nw*Nu))==NULL ||
	    (Q=(double **)malloc(sizeof(double *)*(Nu+1)))==NULL ||
	    (*Q=(double *)malloc(sizeof(double)*((size_t)Nu*Nw+1)))==NULL ||
	    (S4BP=(Struct **)malloc(sizeof(Struct *)*N))==NULL ||
	    (*S4BP=(Struct *)malloc(sizeof(Struct)*N*N))==NULL ||
	    (F=(Float ***)malloc(sizeof(Float **)*N))==NULL ||
	    (*F=(Float **)malloc(sizeof(Float *)*N*N))==NULL ||
	    (**F=(Float *)malloc(sizeof(Float)*N*N*Nl))==NULL)
	    Error("memory allocation error.");

	for (j=0; j<N2; j++) {
	    sincos(M_PI*(double)j/(double)N2,&(E[j].y),&(E[j].x));

	    G[j].x=Filter(j-N1); G[j].y=0.0;
	}
	FFT(-1,G);

	for (i=1; i<Threads; i++) V[i]=V[i-1]+N2;

	for (l=0; l<Nw; l++) {
	    if (l>0) {
		F4P[l]=F4P[l-1]+Nu; P[l]=P[l-1]+Nu;
	    }
	    w=Dw*((double)l-Ow);
	    for (j=0; j<Nu; j++) {
		u=Du*((double)j-Ou);
		F4P[l][j]=A/Du/sqrt(B*B+u*u+w*w)/(double)N2;
	    }
	}
	Q[Nu]=Q[0]; for (j=1; j<Nu; j++) Q[j]=Q[j-1]+Nw;

	O2=O*O;
	for (Y=(-O), y=0; y<N; y++, Y++) {
	    if (y>0) {
		S4BP[y]=S4BP[y-1]+N; F[y]=F[y-1]+N;
	    }
	    O2_Y2=O2-Y*Y;
	    for (X=(-O), x=0; x<N; x++, X++) {
		S4BP[y][x].j=(O2_Y2>=X*X)-1;

		if (y>0 || x>0) F[y][x]=F[y][x-1]+Nl;

		for (z=0; z<Nl; z++) F[y][x][z]=0.0;
	    }
	}
	for (m=0; m<Nt; m++) {
	    if (fread(*P,sizeof(double),(size_t)Nw*Nu,file)!=(size_t)Nw*Nu)
		Error("projection data read error.");

	    if (m>0)
		for (i=0; i<Threads; i++) TERM_MT(Thread[i],BackProjection);

	    for (i=0; i<Threads; i++) INIT_MT(Thread[i],Convolution,i);

	    sincos(t0+2.0*M_PI*(double)m/(double)Nt,&s,&c);
	    for (Y=(-O), y=0; y<N; y++, Y++)
	    for (X=(-O), x=0; x<N; x++, X++)
		if (S4BP[y][x].j>=0) {
		    S4BP[y][x].F4l=F4l=B/(B+Du*(c*(double)Y-s*(double)X));
		    S4BP[y][x].F4Q=F4l*F4l*B*B*M_PI/A/A/(double)Nt;
		    S4BP[y][x].j=j=(int)(u=Ou+F4l*(c*(double)X+s*(double)Y));
		    S4BP[y][x].Eu=u-(double)j;
		}

	    for (i=0; i<Threads; i++) TERM_MT(Thread[i],Convolution);
	    for (i=0; i<Threads; i++) INIT_MT(Thread[i],BackProjection,i);
	}
	for (i=0; i<Threads; i++) TERM_MT(Thread[i],BackProjection);

	if (file!=stdin) (void)fclose(file);

	z=1; for (l=Nw-1; l>=10; l/=10) ++z; (void)sprintf(form,"%%0%dd",z);

	lim=(1<<BPS)-1;

	if (argc==13)
	    (void)sprintf(range,"%lf\t%lf",base,roof=base+step*(double)lim);
	else if (norm) {
	    min=max=F[0][0][0];
	    for (y=0; y<N ; y++)
	    for (x=0; x<N ; x++)
	    for (z=0; z<Nl; z++)
		if (F[y][x][z]<min) min=F[y][x][z]; else
		if (F[y][x][z]>max) max=F[y][x][z];

	    base=min; step=(min<max)?(max-min)/(double)lim:1.0;
	    (void)sprintf(range,"%lf\t%lf",min,max);
	    (void)fprintf(stderr,"%s\n",range);
	}
	for (z=0; z<Nl; z++) {
	    min=max=F[0][0][z];
	    for (y=0; y<N; y++)
	    for (x=0; x<N; x++)
		if (F[y][x][z]<min) min=F[y][x][z]; else
		if (F[y][x][z]>max) max=F[y][x][z];

	    (void)sprintf(slice,form,z+Ol);
	    (void)printf("%s\t%d\t%lf\t%lf\n",slice,O,min,max);

	    if (argc==13)
		for (y=0; y<N; y++) {
		    cell=(Cell *)S4BP[N-1-y];
		    for (x=0; x<N; x++)
			cell[x]=(F[y][x][z]<base)?0:
				(F[y][x][z]>roof)?(Cell)lim:
				(Cell)((F[y][x][z]-base)/step+0.5);
		}
	    else {
		if (!norm) {
		    base=min; step=(min<max)?(max-min)/(double)lim:1.0;
		    (void)sprintf(range,"%lf\t%lf",min,max);
		}
		for (y=0; y<N; y++) {
		    cell=(Cell *)S4BP[N-1-y];
		    for (x=0; x<N; x++)
			cell[x]=(Cell)((F[y][x][z]-base)/step+0.5);
		}
	    }
	    (void)sprintf(str,"%s/%s.tif",argv[argc-1],slice);
	    StoreImageFile(str,N,N,BPS,(Cell **)S4BP,range);
	}
	free(**F); free(*F); free(F);
	free(*S4BP); free(S4BP); free(*Q); free(Q);
	free(*P); free(P); free(*F4P); free(F4P);
	free(*V); free(V); free(G); free(E);
	free(Thread);

	return 0;
}
